#ifndef __MYBITSTRING__
#define __MYBITSTRING__

class myBitString {
protected:
  int size;
  unsigned int *data;
  inline void resize(int newsize)
    {
      unsigned int *olddata;
      if (newsize < size) return;
      //cerr << "Resize " << size << " to " << newsize << endl;
      olddata = data;
      data = new unsigned int[newsize];
      if (!data)
	{
	  cerr << "myBitString.resize(" << newsize << ") out of memory" << endl;
	  exit(1);
	}
      for (int i = 0; i < size; i++)
	data[i] = olddata[i];
      for (int i=size; i< newsize; i++)
	data[i]=0;
      size=newsize;
      if (olddata)
	delete [] olddata;
    };
public:
  inline void optisize(void)
    {
      int i=size-1;
      while (i>=0 && data[i]==0) i--;
      if (i==size-1) return; // nothing to do...

      //cerr << "Optisize " << oldsize << " to " << i+1 << endl;
      if (size==0) { delete [] data; size=0; return; } // now empty

      size=i+1; // new size
      unsigned int *olddata = data;
      data = new unsigned int[size];
      if (!data)
	{
	  cerr << "myBitString.optisize(" << size << ") out of memory" << endl;
	  exit(1);
	}
      for (; i>=0; --i) data[i] = olddata[i];
      if (olddata) delete [] olddata;
    };

  inline myBitString(void)
    {
      size=0;
      data=NULL;
    };
  inline ~myBitString(void)
    {
      if (data) delete [] data;
      data=NULL;
      size=0;
    };
  inline int count(const bool b = true) const
    {
      int i;
      int z=0;
      if (!b) return -1;
      for (i=0; i<size; i++)
	{
	  register unsigned int h = data[i];
	  while (h)
	    {
	      if (h&1) z++;
	      h>>=1;
	    }
	}
      return z;
    }
  inline int first(const bool b = true) const
    {
      if (b)
	{
	  for (int i=0; i<size; i++)
	    {
	      register unsigned int h = data[i];
	      if (h)
		{
		  int j;
		  asm("bsfl %0,%1" : "+g" (h), "=c" (j) );
		  return 32*i+j;
		}
	    }
	  //cerr << "first(1)=-1" << endl;
	  return -1;
	}
      else
	{
	  cerr << "myBitString.first(0) not implemented" << endl;
	  exit(1);
	}
    };
  inline int next(int pos, const bool b = true) const
    {
      int i,j;
      if (pos<0) return first(b);
      pos++;
      if (b)
	{
	  i=pos>>5;
	  j=pos&31;
	  if (i>=size) return -1;
	  register unsigned int h = data[i];
	  h>>=j;
	  if (h)
	    {
	      int j;
	      asm("bsfl %0,%1" : "+g" (h), "=c" (j) );
	      return pos+j;
	    }
	  else
	    {
	      j = 0;
	      for (i++;i<size; i++)
		{
		  h = data[i];
		  if (h)
		    {
		      int j;
		      asm("bsfl %0,%1" : "+g" (h), "=c" (j) );
		      return 32*i+j;
		    }
		}
	    }
	  return -1;
	}
      else
	{
	  cerr << "myBitString.next(0) not implemented" << endl;
	  exit(1);
	}
    };
  inline int last(const bool b = true) const
    {
      int i;

      if (b)
	{
	  for (i=size-1; i>=0; i--)
	    {
	      //cerr << "i=" << i << endl;
	      register unsigned int h = data[i];
	      if (h)
		{
		  int j;
		  asm("bsrl %0,%1" : "+g" (h), "=c" (j) );
		  return 32*i+j;
		}
	    }
	  //cerr << "last(1)=-1"<<endl;
	  return -1;
	}
      else
	{
	  cerr << "myBitString.last(0) not implemented" << endl;
	  exit(1);
	}
    };
  inline int prev(int pos, const bool b = true) const
    {
      int i,j;

      if (pos>32*size) return last(b);
      if (pos<=0) return -1;
      pos--;
      if (b)
	{
	  i=pos>>5;
	  j=pos&31;
	  register unsigned int h = data[i];
	  h<<=31-j;
	  if (h)
	    {
	      int j;
	      asm("bsrl %0,%1" : "+g" (h), "=c" (j) );
	      return pos-(31-j);
	    }
	  else
	    {
	      for (i--;i>=0; i--)
		{
		  h = data[i];
		  if (h)
		    {
		      int j;
		      asm("bsrl %0,%1" : "+g" (h), "=c" (j) );
		      return 32*i+j;
		    }
		}
	    }
	  return -1;
	}
      else
	{
	  cerr << "myBitString.prev(0) not implemented" << endl;
	  exit(1);
	}
    };
  inline void invert(const int pos)
    {
      //cerr << "invert("<<pos<<")"<<endl;
      int i=pos>>5;
      //int j=pos&31;

      if (i>=size) resize(i+1);

      //asm  ("btcl %1,%0" : "+m" (data[i]) : "r" (j));
      asm ("btcl %1,(%0)" : "+q" (data) : "r" (pos));
      //data[i] ^= bitmask[j];
    };
  inline bool test(const int pos) const
    {
      bool r;

      if ((pos>>5)>=size) return false;

      asm ("btl %2,(%1)\n\tsetc %0" : "=r" (r) : "q" (data), "r" (pos));

      //r=(data[i] & bitmask[j])? 1:0;
      return r;
    };
  inline void set(const int pos, const bool b)
    {
      //cerr << "set("<<pos<< "," << b << ")" << endl;
      register int i=pos>>5;
      //int j=pos&31;

      if (i>=size)
	if (b)
	  resize(i+1);
	else
	  return;
      if (b)
	asm  ("btsl %1,(%0)" : "+q" (data) : "r" (pos)); //data[i] |= bitmask[j];
      else
	asm  ("btrl %1,(%0)" : "+q" (data) : "r" (pos)); //data[i] &= ~bitmask[j];
    }
  inline void set(const int pos)
    {
      register int i=pos>>5;

      if (i>=size) resize(i+1);

      asm  ("btsl %1,(%0)" : "+q" (data) : "r" (pos)); //data[i] |= bitmask[j];
    }
  inline void xor(const myBitString &s)
    {
      int i=s.size-1;
      while (i>=0 && s.data[i]==0) i--; // Nullen des zweiten BitStrings ignorieren
      if (i>=size) // <-> i+1>size
	{
	  cerr << "myBitString.xor calls resize" << endl;
	  resize(i+1);
	}
      for (; i>=0; --i)
	data[i] ^= s.data[i];
    }
};

#endif
